﻿/*
 * Created by SharpDevelop.
 * User: grobi
 * Date: 24.06.2017
 * Time: 14:12
 * 
 */
using System;
using System.IO.Ports;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace SDcardTest001
{
	class Program
	{
        // change this to false to work in simulation mode, sector data will we read from files located in STORAGE_PATH
        const bool COMDISABLED = false;
        // the serial COM port you want to use
        const string COM = "COM3";
        // change this to false in case you don't want to store the sector data read from the card in STORAGE_PATH when using the "live" serial connection 
        const bool STORE_SECTOR_DATA = true;
        // the path where sector data files will be stored when STORE_SECTOR_DATA is "TRUE", the application will try to create the directory in case it does not exist.
        const string STORAGE_PATH = @".\bin\";

		static SerialPort sPort = null;

        static Dictionary<byte, string> PartitionTypes;
        static void InitPartitionTypes()
        {
            PartitionTypes = new Dictionary<byte, string>();
            PartitionTypes.Add(0x00, "not used");
            PartitionTypes.Add(0x01, "FAT12 ≤ 16MB");
            PartitionTypes.Add(0x04, "FAT16 ≤ 32 MiB");
            PartitionTypes.Add(0x05, "extended");
            PartitionTypes.Add(0x06, "FAT16 > 32 MiB");
            PartitionTypes.Add(0x07, "NTFS / exFAT");
            PartitionTypes.Add(0x0B, "FAT32");
            PartitionTypes.Add(0x0C, "FAT32 LBA");
            PartitionTypes.Add(0x0E, "FAT16 > 32 MiB LBA");
            PartitionTypes.Add(0x0F, "extended LBA");
            PartitionTypes.Add(0x12, "OEM - Partition");
            PartitionTypes.Add(0x27, "Windows RE");
            PartitionTypes.Add(0x42, "dynamic");
            PartitionTypes.Add(0x82, "Linux Swap");
            PartitionTypes.Add(0x83, "Linux Native");
            PartitionTypes.Add(0x8E, "Linux LVM");
            PartitionTypes.Add(0xA5, "FreeBSD");
            PartitionTypes.Add(0xA6, "OpenBSD");
            PartitionTypes.Add(0xA9, "NetBSD");
            PartitionTypes.Add(0xEE, "Legacy MBR + EFI Header");
            PartitionTypes.Add(0xEF, "EFI filesystem");
        }

        struct PartitionTabelEntry
        {
            public bool Active;
            public UInt32 CHS_first_sector;
            public byte PartitionType;
            public UInt32 CHS_last_sector;
            public UInt32 LBA_start_sector;
            public UInt32 LBA_sectors;
        }
        static PartitionTabelEntry[] PartitionTable;

        struct DirectoryEntry
        {
            public string Name;
            public List<UInt32> Files;
            public List<UInt32> Directories;
            public UInt32 ParentDirectory;
            public UInt32 Cluster;
            public bool Initialized;
        }

        struct FileEntry
        {
            public string Name;
            public UInt32 Directory;
            public UInt32 Cluster;
            public UInt32 Size;
        }

        static Dictionary<UInt32, DirectoryEntry> directories;

        static Dictionary<UInt32, FileEntry> files;

        static DirectoryEntry rootDirectory;

        static UInt16 BytesPerSector = 512;
        static byte SectorsPerCluster = 8;
        static UInt16 ReservedSectors = 0;
        static byte NumberOfFATs = 2;
        static UInt32 SectorsPerFAT = 0;
        static UInt32 RootDirCluster = 2;
        static UInt32 FAT_start_LBA = 0;
        static UInt32 FAT_data_start_LBA = 0;
        static UInt32 ClusterEntriesInFATCluster = 1024;

        static Dictionary<UInt32, byte[]> FAT_sector_data;

        public static void Main(string[] args)
		{
			Console.WriteLine("SD card Test001 by Grobi");
            InitPartitionTypes();
            FAT_sector_data = new Dictionary<UInt32, byte[]>();
            directories = new Dictionary<uint, DirectoryEntry>();
            files = new Dictionary<uint, FileEntry>();
            if (!Directory.Exists(STORAGE_PATH))
            {
                Directory.CreateDirectory(STORAGE_PATH);
            }
            try
			{
                if (!COMDISABLED)
                {
                    // init COM port when needed
                    Console.Write("Trying to open COM port " + COM + " ... ");
                    sPort = new SerialPort(COM, 115200, Parity.None, 8, StopBits.One);
                    //sPort.ReadTimeout = -1;
                    sPort.ReadTimeout = 5000;
                    sPort.Open();
                    Console.WriteLine("OK");
                }
                else Console.WriteLine("Serial COM Port is not used!");

                // read the partition table from the MBR (sector 0)
                Console.Write("Reading the Partition table ... ");
                GetPartitionTable();
                Console.WriteLine("DONE");

                // show found partitions and select first FAT32 partition we can find
                Console.WriteLine("Looking for FAT32 type partition ... ");
                int partitionIndex = -1;
                for (int i = 0; i < 4; i++)
                {
                    Console.WriteLine("Partition " + i.ToString() + " :");
                    string text = String.Empty;
                    if ((partitionIndex == -1) & ((PartitionTable[i].PartitionType == 0x0B) | (PartitionTable[i].PartitionType == 0x0C))) partitionIndex = i;
                    if (PartitionTypes.ContainsKey(PartitionTable[i].PartitionType)) text += PartitionTypes[PartitionTable[i].PartitionType] + " ";
                    else
                    {
                        Console.WriteLine("unknown");
                        continue;
                    }
                    if (PartitionTable[i].Active) text += "A ";
                    text += "0x" + PartitionTable[i].CHS_first_sector.ToString("X6") + " ";
                    text += "0x" + PartitionTable[i].CHS_last_sector.ToString("X6") + " ";
                    text += "0x" + PartitionTable[i].LBA_start_sector.ToString("X8") + " ";
                    text += PartitionTable[i].LBA_sectors.ToString();
                    Console.WriteLine(text);
                }
                if (partitionIndex == -1) throw new Exception("No useable FAT32 partition found!");
                else Console.WriteLine("selecting Partition " + partitionIndex.ToString());

                // init partition
                Console.WriteLine("Init Partition " + partitionIndex.ToString() + " ... ");
                InitPartition(partitionIndex);
                Console.WriteLine("Bytes per sector: " + BytesPerSector.ToString());
                Console.WriteLine("Sectors per cluster: " + SectorsPerCluster.ToString());
                Console.WriteLine("Number of reserved sectors: " + ReservedSectors.ToString());
                Console.WriteLine("Number of FATs: " + NumberOfFATs.ToString());
                Console.WriteLine("Sectors per FAT: " + SectorsPerFAT.ToString());
                Console.WriteLine("Cluster number of the start of the Root Directory: " + RootDirCluster.ToString());
                Console.WriteLine("FAT start LBA: 0x" + FAT_start_LBA.ToString("X8"));
                Console.WriteLine("directory and data clusters start LBA: 0x" + FAT_data_start_LBA.ToString("X8"));
                Console.WriteLine("DONE");

                // read the root directory
                Console.Write("Parsing the Root directory ... ");
                rootDirectory = new DirectoryEntry();
                rootDirectory.Name = string.Empty;
                rootDirectory.ParentDirectory = 0;
                rootDirectory.Cluster = RootDirCluster;
                rootDirectory.Files = new List<UInt32>();
                rootDirectory.Directories = new List<UInt32>();
                rootDirectory.Initialized = false;

                ReadDirectory(ref rootDirectory);
                directories.Add(rootDirectory.Cluster, rootDirectory);
                Console.WriteLine("DONE");

                // show the help.
                ShowHelp();

                // terminal mode
                UInt32 currentdir = rootDirectory.Cluster;
                string diskPath = string.Empty;
                bool exitLoop = false;
                while (!exitLoop)
                {
                    Console.Write(diskPath + ">");
                    string input = Console.ReadLine().Trim();
                    string command = string.Empty;
                    string param = string.Empty;
                    int z = input.IndexOf(' ');
                    if (z != -1)
                    {
                        command = input.Substring(0, z).ToUpper();
                        param = input.Substring(z + 1).Trim();
                    }
                    else command = input.ToUpper();
                    switch(command)
                    {
                        case "EXIT":
                            {
                                exitLoop = true;
                                break;
                            }
                        case "LS":
                            {
                                ListDirectory(directories[currentdir]);
                                break;
                            }
                        case "CD":
                            {
                                if (!string.IsNullOrEmpty(param))
                                {
                                    string dirName = param;
                                    if ((dirName != ".") && (dirName != ".."))
                                    {
                                        DirectoryEntry de = directories[currentdir];
                                        int i = 0;
                                        for (i = 0; i < de.Directories.Count; i++)
                                            if (directories[de.Directories[i]].Name == dirName) break;
                                        if (i < de.Directories.Count)
                                        {
                                            currentdir = de.Directories[i];
                                            if (directories[currentdir].Initialized == false)
                                            {
                                                DirectoryEntry newDe = directories[currentdir];
                                                ReadDirectory(ref newDe);
                                                directories[currentdir] = newDe;
                                            }
                                            if (string.IsNullOrEmpty(diskPath)) diskPath = directories[currentdir].Name;
                                            else diskPath += "\\" + directories[currentdir].Name;
                                        }
                                        else Console.WriteLine("Directory " + dirName + " not found!");
                                    }
                                    else if (dirName == "..")
                                    {
                                        currentdir = directories[currentdir].ParentDirectory;
                                        if (currentdir == 0) currentdir = rootDirectory.Cluster;
                                        if (!string.IsNullOrEmpty(diskPath))
                                        {
                                            int i = diskPath.LastIndexOf('\\');
                                            if (i > -1) diskPath = diskPath.Substring(0, i - 1);
                                            else diskPath = string.Empty;    
                                        }
                                    }
                                }
                                else Console.WriteLine("Incomplete command!");
                                break;
                            }
                        case "CAT":
                            {
                                if (!string.IsNullOrEmpty(param))
                                {
                                    string fileName = param;
                                    DirectoryEntry de = directories[currentdir];
                                    int i = 0;
                                    for (i = 0; i < de.Files.Count; i++)
                                        if (files[de.Files[i]].Name == fileName) break;
                                    if (i < de.Files.Count)
                                    {
                                        FileEntry fe = files[de.Files[i]];
                                        byte[] data = ReadFile(ref fe);
                                        string textData = ASCIIEncoding.UTF8.GetString(data);
                                        Console.WriteLine(textData);
                                    }
                                    else Console.WriteLine("File " + fileName + " not found!");
                                }
                                else Console.WriteLine("Incomplete command!");
                                break;
                            }
                        case "DL":
                            {
                                if (!string.IsNullOrEmpty(param))
                                {
                                    string fileName = param;
                                    DirectoryEntry de = directories[currentdir];
                                    int i = 0;
                                    for (i = 0; i < de.Files.Count; i++)
                                        if (files[de.Files[i]].Name == fileName) break;
                                    if (i < de.Files.Count)
                                    {
                                        FileEntry fe = files[de.Files[i]];
                                        byte[] data = ReadFile(ref fe);
                                        Console.Write("Enter filename to store on disk: ");
                                        string fn = Console.ReadLine();
                                        try
                                        {
                                            FileStream fs = new FileStream(fn, FileMode.Create);
                                            BinaryWriter bw = new BinaryWriter(fs);
                                            bw.Write(data, 0, data.Length);
                                            bw.Close();
                                            fs.Close();
                                            Console.WriteLine("DONE");
                                        }
                                        catch(Exception ex)
                                        {
                                            Console.WriteLine("Error saving data to disk!" + Environment.NewLine + ex.Message);
                                        }
                                    }
                                    else Console.WriteLine("File " + fileName + " not found!");
                                }
                                else Console.WriteLine("Incomplete command!");
                                break;
                            }
                        case "HELP":
                            {
                                ShowHelp();
                                break;
                            }
                        default:
                            {
                                Console.WriteLine("Unknown command!");
                                break;
                            }
                    }
                    System.Threading.Thread.Sleep(1);
                }
                Console.WriteLine();
			}
			catch(Exception ex)
			{
				Console.WriteLine("[ERROR]: " + ex.Message);
			}
            if ((!COMDISABLED) & (sPort != null))
            {
                if (sPort.IsOpen) sPort.Close();
            }
			Console.Write("Press any key to continue . . . ");
			Console.ReadKey(true);
		}

        static void ShowHelp()
        {
            Console.WriteLine(Environment.NewLine + "Help:");
            Console.WriteLine("ls                   list the directory content");
            Console.WriteLine("cd directory-name    change the directory");
            Console.WriteLine("cat file-name        shows the content of the file");
            Console.WriteLine("dl file-name         downloads the file");
            Console.WriteLine("help                 shows this help text");
            Console.WriteLine("exit                 quit");
            Console.WriteLine();
        }

        static void ListDirectory(DirectoryEntry dir)
        {
            List<string> _directories = new List<string>();
            List<string> _files = new List<string>();
            foreach(UInt32 dc in dir.Directories)
            {
                DirectoryEntry de = directories[dc];
                string text = de.Name + "\t[D]\t(";
                if ((de.Directories.Count > 0) | (de.Files.Count > 0)) text += Convert.ToString(de.Directories.Count + de.Files.Count) + ")";
                else text += "?)";
                _directories.Add(text);
            }
            foreach (UInt32 fc in dir.Files)
            {
                FileEntry fe = files[fc];
                string text = fe.Name + "\t[A]\t" + fe.Size.ToString() + " byte(s)";
                _files.Add(text);
            }
             _directories.Sort();
            _files.Sort();
            for (int i = 0; i < _directories.Count; i++)
                Console.WriteLine(_directories[i]);
            for (int i = 0; i < _files.Count; i++)
                Console.WriteLine(_files[i]);
            Console.WriteLine(Convert.ToString(_directories.Count + _files.Count) + " item(s) total.");
        }

        static string CleanShortFilename(string shortName)
        {
            string result = string.Empty;
            string ext = shortName.Substring(8, 3).Trim();
            result = shortName.Substring(0, 8).Trim();
            if (!string.IsNullOrEmpty(ext)) result += "." + ext;
            return result;
        }

        static void GetPartitionTable()
        {
            byte[] MBR = ReadSector(0);
            UInt16 MagicNumber = (UInt16)((MBR[510] << 8) + MBR[511]);
            if (MagicNumber != 0x55AA) throw new Exception("MBR MagicNumber mismatch!");
            PartitionTable = new PartitionTabelEntry[4];
            int addr = 446;
            for (int i = 0; i < 4; i++)
            {
                PartitionTabelEntry entry = new PartitionTabelEntry();
                if (MBR[addr] == 0x80) entry.Active = true;
                else entry.Active = false;
                entry.CHS_first_sector = (UInt32)(((MBR[addr + 1] << 16) + (MBR[addr + 2] << 8) + MBR[addr + 3]) & 0x00FFFFFF);
                entry.PartitionType = MBR[addr + 4];
                entry.CHS_last_sector = (UInt32)(((MBR[addr + 5] << 16) + (MBR[addr + 6] << 8) + MBR[addr + 7]) & 0x00FFFFFF);
                entry.LBA_start_sector = BitConverter.ToUInt32(MBR, addr + 8);
                entry.LBA_sectors = BitConverter.ToUInt32(MBR, addr + 12);
                PartitionTable[i] = entry;
                addr += 16;
            }
        }

        static void InitPartition(int partitionIndex)
        {
            if ((PartitionTable == null) || (PartitionTable.Length <= partitionIndex)) throw new Exception("No partition data found at index " + partitionIndex.ToString() + " !");
            byte[] partbootrecord = ReadSector(BytesPerSector * PartitionTable[partitionIndex].LBA_start_sector);
            UInt32 MagicNumber = (UInt16)((partbootrecord[510] << 8) + partbootrecord[511]);
            if (MagicNumber != 0x55AA) throw new Exception("FAT MagicNumber mismatch!");
            BytesPerSector = (UInt16)((partbootrecord[12] << 8) + partbootrecord[11]);
            if (BytesPerSector != 512) throw new Exception("We expected 512 bytes per sector!");
            SectorsPerCluster = partbootrecord[13];
            ReservedSectors = (UInt16)((partbootrecord[15] << 8) + partbootrecord[14]);
            NumberOfFATs = partbootrecord[16];
            SectorsPerFAT = (UInt32)((partbootrecord[39] << 24) + (partbootrecord[38] << 16) + (partbootrecord[37] << 8) + partbootrecord[36]);
            RootDirCluster = (UInt32)((partbootrecord[47] << 24) + (partbootrecord[46] << 16) + (partbootrecord[45] << 8) + partbootrecord[44]);
            FAT_start_LBA = PartitionTable[partitionIndex].LBA_start_sector + ReservedSectors;
            FAT_data_start_LBA = PartitionTable[partitionIndex].LBA_start_sector + ReservedSectors + (NumberOfFATs * SectorsPerFAT);
            ClusterEntriesInFATCluster = (UInt32)((BytesPerSector / 4) * SectorsPerCluster);
        }

        static UInt32 GetNextCluster(UInt32 cluster)
        {
            UInt32 fat_cluster = cluster / ClusterEntriesInFATCluster;
            UInt32 fat_cluster_entry = cluster % ClusterEntriesInFATCluster;
            UInt32 fat_cluster_sector = (fat_cluster_entry * 4) / BytesPerSector;
            int addr = (int)((fat_cluster_entry * 4) - (fat_cluster_sector * BytesPerSector));
            UInt32 fat_sector_index = fat_cluster * SectorsPerCluster + fat_cluster_sector;
            byte[] fat_data = null;
            if (FAT_sector_data.ContainsKey(fat_sector_index)) fat_data = FAT_sector_data[fat_sector_index];
            else
            {
                UInt32 fat_addr_lba = FAT_start_LBA + (fat_cluster * SectorsPerCluster) + fat_cluster_sector;
                fat_data = ReadSector(fat_addr_lba * BytesPerSector);
                FAT_sector_data.Add(fat_sector_index, fat_data);
            }
            UInt32 nextCluster = BitConverter.ToUInt32(fat_data, addr);
            return nextCluster;
        }

        static void ParseDirectoryDataSector(byte[] dir, ref DirectoryEntry directory, ref bool lastSector)
        {
            int addr = 0;
            Dictionary<int, string> lfn_parts = new Dictionary<int, string>();
            bool lfn = false;
            while ((addr < 512) && (dir[addr] != 0x00))
            {
                if (dir[addr] != 0xE5)
                {
                    string shortName = CleanShortFilename(ASCIIEncoding.UTF8.GetString(dir, addr, 11));
                    byte attr = dir[addr + 11];
                    UInt16 FirstClusterHigh = BitConverter.ToUInt16(dir, addr + 20);
                    UInt16 FirstClusterLow = BitConverter.ToUInt16(dir, addr + 26);
                    UInt32 FileSize = BitConverter.ToUInt32(dir, addr + 28);
                    UInt32 clusterNumber = (UInt32)((FirstClusterHigh << 16 | FirstClusterLow) & 0xFFFFFFFF);
                    if ((attr & 0x0F) == 0x0F)
                    {
                        // long file name
                        lfn = true;
                        string longName = string.Empty;
                        longName = ASCIIEncoding.Unicode.GetString(dir, addr + 1, 10);
                        longName += ASCIIEncoding.Unicode.GetString(dir, addr + 14, 12);
                        longName += ASCIIEncoding.Unicode.GetString(dir, addr + 28, 4);
                        int temp = longName.IndexOf('\0');
                        if (temp != -1) longName = longName.Substring(0, temp);
                        temp = dir[addr];
                        if ((temp & 0x40) == 0x40) temp -= 0x40;
                        if (!lfn_parts.ContainsKey(temp)) lfn_parts.Add(temp, longName);
                        else throw new Exception("Long FileName part is already in our list!");
                    }
                    else if ((attr & 0x10) == 0x10)
                    {
                        DirectoryEntry newDir = new DirectoryEntry();
                        newDir.Name = shortName;
                        newDir.Files = new List<UInt32>();
                        newDir.Directories = new List<UInt32>();
                        newDir.ParentDirectory = directory.Cluster;
                        newDir.Cluster = clusterNumber;
                        if (lfn)
                        {
                            string longName = string.Empty;
                            for (int i = 1; i <= lfn_parts.Count; i++)
                                if (lfn_parts.ContainsKey(i)) longName += lfn_parts[i];
                                else throw new Exception("Missing Long Filename part!");
                            lfn_parts.Clear();
                            lfn = false;
                            newDir.Name = longName;
                        }
                        if ((newDir.Name != ".") && (newDir.Name != ".."))
                        {
                            directory.Directories.Add(newDir.Cluster);
                            directories.Add(newDir.Cluster, newDir);
                        }
                    }
                    else if ((attr & 0x20) == 0x20)
                    {
                        FileEntry newFile = new FileEntry();
                        newFile.Name = shortName;
                        newFile.Directory = directory.Cluster;
                        newFile.Cluster = clusterNumber;
                        newFile.Size = FileSize;
                        if (lfn)
                        {
                            string longName = string.Empty;
                            for (int i = 1; i <= lfn_parts.Count; i++)
                                if (lfn_parts.ContainsKey(i)) longName += lfn_parts[i];
                                else throw new Exception("Missing Long Filename part!");
                            lfn_parts.Clear();
                            lfn = false;
                            newFile.Name = longName;
                        }
                        directory.Files.Add(newFile.Cluster);
                        files.Add(newFile.Cluster, newFile);
                    }
                }
                addr += 32;
            }
            if ((addr < 512) && (dir[addr] == 0x00)) lastSector = true;
            else lastSector = false;
        }

        static void ReadDirectory(ref DirectoryEntry directory)
        {
            UInt32 currentCluster = directory.Cluster;
            UInt32 dir_start_LBA = 0;
            UInt32 nextCluster = 0;
            bool lastSector = false;
            byte[] dirData = null;
            do
            {
                lastSector = false;
                dir_start_LBA = FAT_data_start_LBA + (currentCluster - 2) * SectorsPerCluster;
                for (int i = 0; (i < SectorsPerCluster) && (lastSector == false); i++)
                {
                    dirData = ReadSector((UInt32)(dir_start_LBA + i) * BytesPerSector);
                    ParseDirectoryDataSector(dirData, ref directory, ref lastSector);
                }
                if (!lastSector)
                {
                    nextCluster = GetNextCluster(currentCluster);
                    if (nextCluster != 0x0FFFFFFF) currentCluster = nextCluster;
                }
            }
            while ((nextCluster != 0x0FFFFFFF) && (lastSector == false));
            directory.Initialized = true;
        }

        static byte[] ReadFile(ref FileEntry file)
        {
            List<byte> data = new List<byte>();
            UInt32 currentCluster = file.Cluster;
            UInt32 file_start_LBA = 0;
            UInt32 nextCluster = 0;
            UInt32 bytesToRead = file.Size;
            byte[] fileData = null;
            do
            {
                file_start_LBA = FAT_data_start_LBA + (currentCluster - 2) * SectorsPerCluster;
                for (int i = 0; (i < SectorsPerCluster) && (bytesToRead > 0); i++)
                {
                    fileData = ReadSector((UInt32)(file_start_LBA + i) * BytesPerSector);
                    if (bytesToRead >= fileData.Length)
                    {
                        data.AddRange(fileData);
                        bytesToRead -= (UInt32)fileData.Length;
                    }
                    else
                    {
                        byte[] temp = new byte[bytesToRead];
                        Array.Copy(fileData, temp, bytesToRead);
                        data.AddRange(temp);
                        bytesToRead -= (UInt32)temp.Length;
                    }
                }
                if (bytesToRead > 0)
                {
                    nextCluster = GetNextCluster(currentCluster);
                    if (nextCluster != 0x0FFFFFFF) currentCluster = nextCluster;
                }
            } while ((bytesToRead > 0) && (nextCluster != 0x0FFFFFFF));
            return data.ToArray();
        }

        static byte[] ReadSector(UInt32 addr)
        {
            if (COMDISABLED) return ReadSectorFromFile(addr);
            FileStream fs = null;
            BinaryWriter bw = null;
            byte[] command = new byte[5] { 0x52, (byte)((addr >> 24) & 0xFF), (byte)((addr >> 16) & 0xFF), (byte)((addr >> 8) & 0xFF), (byte)(addr & 0xFF)};
            byte[] data = new byte[512];
            if (STORE_SECTOR_DATA == true)
            {
                string filename = STORAGE_PATH + addr.ToString("X8") + ".bin";
                fs = new FileStream(filename, FileMode.Create);
                bw = new BinaryWriter(fs);
            }
            sPort.Write(command, 0,  command.Length);
            int rb = 0;
            while (rb < 512)
            {
                rb += sPort.Read(data, rb, 512 - rb);
                System.Threading.Thread.Sleep(1);
            }
            // skip checksum
            //sPort.ReadByte();
            //sPort.ReadByte();
            if (STORE_SECTOR_DATA)
            {
                bw.Write(data, 0, data.Length);
                bw.Close();
                fs.Close();
            }
            return data;
        }

        static byte[] ReadSectorFromFile(UInt32 addr)
        {
            FileStream fs = null;
            BinaryReader br = null;
            byte[] data = new byte[512];
            string filename = STORAGE_PATH + addr.ToString("X8") + ".bin";
            if (!File.Exists(filename)) throw new Exception("File " + filename + " does not exist!");
            fs = new FileStream(filename, FileMode.Open);
            br = new BinaryReader(fs);
            data = br.ReadBytes(512);
            br.Close();
            fs.Close();
            return data;
        }

    }
}